import random
import re
import string
from typing import Any, Callable

from .visitor import Result

USER_AGENT = "https://github.com/am230/iwashi"
BASE_HEADERS = {"User-Agent": USER_AGENT}
HTTP_REGEX = "(https?://)?(www.)?"


def print_result(
    result: Result, indent_level=0, print: Callable[[str], Any] = print
) -> None:
    indent = indent_level * "    "
    print(f"{indent}{result.site_name}")
    print(f"{indent}│url  : {result.url}")
    print(f"{indent}│name : {result.title}")
    print(f"{indent}│score: {result.score}")
    print(f"{indent}│links : {result.links}")
    if result.description:
        print(f"{indent}│description: " + result.description.replace("\n", "\\n"))
    for child in result.children:
        print_result(child, indent_level + 1, print)


def parse_host(url: str) -> str:
    match = re.search(r"(https?:\/\/)?(www\.)?(?P<host>[\w.]+)/", url)
    if match is None:
        return url
    return match.group("host")


def random_string(length: int) -> str:
    return "".join(random.choice(string.ascii_letters) for _ in range(length))


URL_NORMALIZE_REGEX = r"(?P<protocol>https?)?:?\/?\/?(?P<domain>[^.]+\.[^\/]+)(?P<path>[^?#]+)?(?P<query>.+)?"


def normalize_url(url: str) -> str:
    match = re.match(URL_NORMALIZE_REGEX, url)
    if match is None:
        raise ValueError(f"Invalid URL: {url}")
    return f"{match.group('protocol') or 'https'}://{match.group('domain')}{match.group('path') or ''}{match.group('query') or ''}"



from setuptools import find_packages, setup

with open("README.rst", mode="r", encoding="utf-8") as f:
    long_description = f.read()

author = "am230"
name = "iwashi"
py_modules = [name]

setup(
    name=name,
    version="1.1.1",
    keywords=("generator"),
    description="Link Tree Generator",
    long_description=long_description,
    install_requires=["requests", "bs4", "yt-dlp", "loguru", "typingdict", "click"],
    license="MIT Licence",
    long_description_content_type="text/x-rst",
    packages=find_packages(),
    url=f"https://github.com/{author}/{name}",
    author=author,
    author_email="am.230@outlook.jp",
    py_modules=py_modules,
    platforms="any",
)



from typing import List, MutableSet, Optional

import requests
from loguru import logger

from .helper import BASE_HEADERS, parse_host
from .visitor import Context, Result, SiteVisitor, Visitor


class Iwashi(Visitor):
    def __init__(self) -> None:
        self.visitors: List[SiteVisitor] = []
        self.visited: MutableSet[str] = set()

    def add_visitor(self, visitor: SiteVisitor) -> None:
        self.visitors.append(visitor)

    def is_visited(self, url: str) -> bool:
        return url in self.visited

    def mark_visited(self, url: str) -> bool:
        url = url.lower()
        if self.is_visited(url):
            return False
        self.visited.add(url)
        return True

    def visit(self, url: str, context: Optional[Context] = None) -> Optional[Result]:
        context = context or Context(url=url, visitor=self)
        if self.is_visited(url):
            return None
        for visitor in self.visitors:
            match = visitor.match(url, context)
            if match is None:
                continue

            normalized = visitor.normalize(url)
            if self.mark_visited(normalized):
                match = visitor.match(normalized, context)
                if match is not None:
                    visitor.visit(normalized, context, **match.groupdict())
            elif context.parent is not None:
                context.parent.link(normalized)
            break
        else:
            self.try_redirect(url, context)
            context.create_result(site_name=parse_host(url), url=url, score=1.0)
            self.mark_visited(url)
            logger.warning(f"[No Visitor Found] {url}")

        return context.result

    def try_redirect(self, url: str, context: Context) -> None:
        res = requests.get(url, allow_redirects=True, headers=BASE_HEADERS)
        if res.url == url:
            return
        context.visit(res.url)
        logger.info(f"[Redirect] {url} -> {res.url}")


def get_iwashi():
    iwashi = Iwashi()
    from . import visitors

    for attr in dir(visitors):
        value = getattr(visitors, attr)
        if attr.startswith("_"):
            continue
        if isinstance(value, type) and issubclass(value, SiteVisitor):
            iwashi.add_visitor(value())
    return iwashi


def visit(url: str, iwashi: Optional[Iwashi] = None) -> Optional[Result]:
    iwashi = iwashi or get_iwashi()
    return iwashi.visit(url)
